package be.iminds.ilabt.jfed.experimenter_gui.util;

import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Resource;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ResourceClass;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.TestbedCategory;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.util.jsonld.impl.PrimaryIdObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;

public class LinkInfoMatcher {
    private static final Logger LOG = LoggerFactory.getLogger(LinkInfoMatcher.class);

    private LinkInfoMatcher() { }

    @Nonnull
    private static Resource.LinkInfoRules legacyLinkInfoRules(@Nullable Resource fromResource, @Nullable Resource toResource) {
        List<Resource> resources = new ArrayList<>();
        if (fromResource != null) resources.add(fromResource);
        if (toResource != null) resources.add(toResource);
        return legacyLinkInfoRules(resources);
    }

    private static Resource.LinkInfoRules legacyLinkInfoRules(@Nonnull List<Resource> resources) {
        //Note: if legacy stuff is removed, this can be replaced by a fallback that refuses any link.

        if (resources.isEmpty())
            return new Resource.LinkInfoRules(false, false, Collections.emptyList(), null);

        Resource firstResource = resources.get(0);
        assert firstResource != null;

        //use legacy stuff to construct fallback LinkInfo
        boolean stitching = false;
        Integer firstServerId = firstResource.getServerId();
        for (Resource r: resources) {
            assert r != null;
            if (r.getServerId() != null && firstServerId == null)
                firstServerId = r.getServerId();
            if (r.getServerId() != null && !Objects.equals(r.getServerId(), firstServerId))
                stitching = true;
        }

        if (!stitching && firstResource.getAllowLinks() != null) {
            return new Resource.LinkInfoRules(firstResource.getAllowLinks(), true,
                    firstResource.getAllowedLinkTypes(), firstResource.getDefaultLinkType());
        }
        if (stitching) {
            boolean use = false;
            Resource usedR = null;
            for (Resource r: resources) {
                if (r.getAllowStitchedLinks() != null) {
                    use = true;
                    if (usedR == null) usedR = r;
                }
                //allow takes priority if resources disagree
                if (r.getAllowStitchedLinks() == Boolean.TRUE) {
                    usedR = r;
                }
            }
            if (use) {
                return new Resource.LinkInfoRules(usedR.getAllowStitchedLinks() == Boolean.TRUE, true,
                        usedR.getAllowedLinkTypes(), usedR.getDefaultLinkType());
            }
        }

        //use even older legacy
        if (!stitching) {
            //falls back to not allowing link when nothing specified
            return new Resource.LinkInfoRules(firstResource.getServer().getTestbed().getAllowLinks() == Boolean.TRUE,
                    true, Collections.singletonList("lan"), null);
        } else {
            //legacy: if any involved server has stitching flag, assume stitching is OK
            for (Resource r: resources) {
                if (r.getServer().hasFlag(Server.Flag.featureStitching))
                    return new Resource.LinkInfoRules(true,
                            false, Collections.singletonList("lan"), "lan");
            }
            //fall back to not allowing
            return new Resource.LinkInfoRules(false,
                    false, Collections.singletonList("lan"), null);
        }
    }

    /**
     *
     * @param config
     * @param nodes the nodes that the link connection. Begin with "from" node(s) if relevant.
     * @return
     */
    @Nonnull
    public static Resource.LinkInfoRules findLinkInfoRules(@Nonnull JFedGuiConfig config, @Nonnull List<FXRspecNode> nodes) {
        if (nodes.isEmpty())
            return new Resource.LinkInfoRules(false, false, Collections.emptyList(), null);

        if (nodes.size() == 1)
            return new Resource.LinkInfoRules(false, false, Collections.emptyList(), null);

        FXRspecNode firstNode = nodes.get(0);
        assert firstNode != null;

        List<Resource> allResources = new ArrayList<>(nodes.size());
        List<Resource> uniqueResources = new ArrayList<>(nodes.size());
        for (FXRspecNode node: nodes) {
            @Nullable ResourceClass nodeResourceClass = node.getResourceClassId() == null ? null : config.getResourceClass(node.getResourceClassId());
            Resource nodeResource = config.findBestResource(nodeResourceClass, node, null);
            if (nodeResource == null) {
                LOG.warn("findLinkInfoRules Could not find Resource for node "+node.getUniqueId()+" -> fallback=refuse all links");
                return new Resource.LinkInfoRules(false, false, Collections.emptyList(), null);
            }
            if (uniqueResources.stream().map(Resource::getId).noneMatch(id -> Objects.equals(id, nodeResource.getId()))) {
                uniqueResources.add(nodeResource);
            }
            allResources.add(nodeResource);
        }

        Resource.LinkInfo bestMatch = null;
        int bestMatchLevel = 0;
        for (int i = 0; i < nodes.size(); i++) {
            FXRspecNode fromNode = nodes.get(i);
            Resource fromResource = allResources.get(i);

            @Nonnull List<Resource.LinkInfo> fromLinkInfos = fromResource.getLinkInfoOrDefault();
            if (fromLinkInfos.isEmpty())
                //no link info, so legacy fallback, unless others have link info
                continue;
            for (int j = 0; j < nodes.size(); j++) {
                if (j == i) continue;
                FXRspecNode toNode = nodes.get(j);
                Resource toResource = allResources.get(j);

                for (Resource.LinkInfo linkInfo: fromLinkInfos) {
                    if (linkInfo.getFilter() != null) {
                        boolean match = false;

                        Resource.LinkInfoFilter filter = linkInfo.getFilter();
                        match |= filter.getMatchSameResource() == Boolean.TRUE && toResource != null && toResource.getId() != null &&
                                Objects.equals(fromResource.getId(), toResource.getId());
                        match |= filter.getMatchAnySameServer() == Boolean.TRUE && toResource != null && toResource.getServerId() != null &&
                                Objects.equals(fromResource.getServerId(), toResource.getServerId());
                        match |= filter.getMatchAnyInterServer() == Boolean.TRUE && toResource != null && toResource.getServerId() != null &&
                                !Objects.equals(fromResource.getServerId(), toResource.getServerId());
                        match |= filter.getMatchResourceId() != null && toResource != null && toResource.getId() != null &&
                                filter.getMatchResourceId().contains(toResource.getId());
                        match |= filter.getMatchServerId() != null && toResource != null && toResource.getServerId() != null &&
                                filter.getMatchServerId().contains(toResource.getServerId());
                        match |= filter.getMatchCategory() != null && toResource != null && toResource.getServerId() != null &&
                                toResource.getServer().getTestbed().getCategories().stream().map(TestbedCategory::getName).anyMatch(cat -> filter.getMatchCategory().contains(cat));
                        match |= filter.getMatchFederation() != null && toResource != null && toResource.getServerId() != null &&
                                toResource.getServer().getTestbed().getFederations().stream().map(PrimaryIdObject::getId).anyMatch(fed -> filter.getMatchFederation().contains(fed));

                        if (filter.getMaxNodeCount() != null && nodes.size() > filter.getMaxNodeCount() ) {
                            match = false;
                        }

                        if (match) {
                            //calculate matchlevel
                            int matchLevel = 0; //default == match on nothing == 0
                            if (filter.getMatchResourceId() != null) matchLevel = 6;
                            if (filter.getMatchServerId() != null) matchLevel = 5;
                            if (filter.getMatchCategory() != null) matchLevel = 4;
                            if (filter.getMatchFederation() != null) matchLevel = 4;
                            if (filter.getMatchSameResource() == Boolean.TRUE) matchLevel = 3;
                            if (filter.getMatchAnySameServer() == Boolean.TRUE) matchLevel = 2;
                            if (filter.getMatchAnyInterServer() == Boolean.TRUE) matchLevel = 1;

                            //note that ORDER of the nodes in the "nodes" argument WILL influence this: for ties, earlier nodes will win (and this is what we want!)
                            if (bestMatch == null || bestMatchLevel < matchLevel) {
                                if (bestMatch != null && bestMatchLevel > 0) {
                                    LOG.warn("Multiple LinkInfo matches for resource "+fromResource.getId()+" overwriting " +
                                            "match with level "+bestMatchLevel+" for match with level "+matchLevel);
                                }
                                bestMatch = linkInfo;
                                bestMatchLevel = matchLevel;
                            } else {
                                LOG.warn("Multiple LinkInfo matches for resource "+fromResource.getId()+" keeping " +
                                        "match with level "+bestMatchLevel+" because new match only has level "+matchLevel);
                            }
                        }
                    } else {
                        if (bestMatch == null) {
                            bestMatch = linkInfo;
                            bestMatchLevel = 0; //default is match level 0
                        }
                    }
                }
            }
        }

        if (bestMatch != null) {
            return bestMatch.getRules();
        }

        //linkInfo provided no match at all, so we fall back to legacy methods
        return legacyLinkInfoRules(allResources);
    }

}
